import math


def format_bytes(size: float) -> str:
    power = 2**10
    n = 0
    power_labels = {0: "bytes", 1: "KB", 2: "MB", 3: "GB", 4: "TB"}

    while size > power:
        size /= power
        n += 1

    return f"{math.ceil(size)} {power_labels[n]}"



import sys

from .cli import main
from .core import Header, Entry  # noqa
from .api import TextureCache, Texture  # noqa
from .find import list_texture_caches  # noqa


if __name__ == "__main__":
    try:
        main()
    except (KeyboardInterrupt, EOFError):
        sys.exit(130)



import argparse
from pathlib import Path
import sys
from typing import Literal
from tqdm import tqdm
import os

from .signal import interrupthandler
from .api import Texture, TextureCache
from .find import find_texturecache, list_texture_caches

OutputMode = Literal["progress", "files", "debug"]


class TextureEmptyError(Exception):
    pass


class TextureIncompleteError(Exception):
    pass


class Args(argparse.Namespace):
    cache_dir: Path | None
    output_dir: Path
    output_mode: OutputMode
    force: bool
    raw: bool
    skip_integrity: bool


def clear_screen() -> None:
    os.system("cls" if os.name == "nt" else "clear")


def prompt_for_cache_dir() -> Path:
    try:
        caches = list_texture_caches()
    except FileNotFoundError:
        print("error: no cache found")
        print('try specificying a cache directory with "texture-courier <cache_dir>"')
        sys.exit(1)

    print("no cache directory specified, enter path or select from the following")
    print("")

    for i, path in enumerate(caches, start=1):
        print(f"{i}: {path.resolve()}")

    print("")

    with interrupthandler(immediate=True) as h:
        while not h.interrupted:
            selection = input("enter path or selection: ")

            if selection.strip() == "":
                continue

            if selection in ("q", "quit", "exit", "0"):
                sys.exit(0)

            if selection.isdigit():
                s = int(selection)

                if s < 1 or s > len(caches):
                    print("invalid selection")
                    continue

                return caches[s - 1]
            else:
                cache = find_texturecache(Path(selection))

                if cache is None:
                    print(f"error: no texture cache found at {selection}")
                    sys.exit(1)

                return cache

    assert False, "unreachable"


def parse_args() -> Args:
    parser = argparse.ArgumentParser(
        prog="texture-courier",
        description="rips texture cache from second life viewers",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument(
        "cache_dir", type=Path, nargs="?", help="path to texture cache directory"
    )
    parser.add_argument(
        "--output-dir",
        "-o",
        type=Path,
        help="path to output extracted textures",
        default="./texturecache",
    )

    parser.add_argument(
        "--output-mode",
        "-O",
        choices=("progress", "files", "debug"),
        help="specify output mode. 'progress' shows a progress bar, 'files' prints the path of each file",
        default="progress",
    )

    parser.add_argument(
        "--watch",
        "-w",
        action="store_true",
        help="watch the cache directory for changes",
        default=False,
    )

    parser.add_argument(
        "--force",
        "-f",
        action="store_true",
        help="overwrite output files",
        default=False,
    )

    parser.add_argument(
        "--raw",
        action="store_true",
        help="skip encoding and just save the raw codestream",
        default=False,
    )

    parser.add_argument(
        "--skip-integrity",
        action="store_true",
        help="skip integrity checks",
        default=False,
    )

    args = Args()
    parser.parse_args(namespace=args)

    return args


def save_texture(texture: Texture, output_dir: Path, args: Args) -> Path:
    if texture.is_empty:
        raise TextureEmptyError

    if texture.is_downloaded() is False and not args.skip_integrity:
        raise TextureIncompleteError

    if args.raw is False:
        save_path = output_dir / f"{texture.uuid}.jp2"

        if save_path.exists() and not args.force:
            raise FileExistsError

        # the cache stores textures in a raw jpeg2000 codestream format
        # that is hard for most operating systems to read, which isn't
        # intended to be used for storage. loading it with pillow puts
        # it in a proper container format
        with texture.open_image() as im:
            im.save(save_path)
    else:
        save_path = output_dir / f"{texture.uuid}.j2c"

        if save_path.exists() and not args.force:
            raise FileExistsError

        save_path.write_bytes(texture.loads())

    # set last access and modification times to the same as the date in cache
    # (atime, mtime)
    os.utime(save_path, (texture.time.timestamp(), texture.time.timestamp()))

    return save_path


def end(
    *,
    args: Args,
    good_writes: int,
    existing_textures: int,
    incomplete_textures: int,
    error_write_textures: int,
    empty_textures: int,
) -> None:
    if args.output_mode in ("progress", "debug"):
        print("")
        print(f"wrote {good_writes} textures to {args.output_dir.resolve()}")
        (
            print(f"skipped {existing_textures} existing textures")
            if existing_textures
            else None
        )
        if incomplete_textures:
            print(f"skipped {incomplete_textures} incomplete textures")
        (
            print(f"{error_write_textures} invalid textures could not be written")
            if error_write_textures
            else None
        )
        print(f"skipped {empty_textures} empty textures") if empty_textures else None
        print("")


def main() -> None:
    args = parse_args()

    if args.cache_dir:
        cache_dir = find_texturecache(args.cache_dir)

        if cache_dir is None:
            print(f"error: no texture cache found at {args.cache_dir.resolve()}")
            sys.exit(1)
    else:
        if args.output_mode == "files":
            print("error: output mode 'files' requires a cache directory")
            sys.exit(1)

        cache_dir = prompt_for_cache_dir()

    cache = TextureCache(cache_dir)
    existing_textures = 0
    good_writes = 0

    if args.output_mode == "debug":
        print("")
        print("TEXTURE ENTRIES HEADER:")

        for k, v in cache.header:
            print(f"{k}: {v}")

    args.output_dir.mkdir(exist_ok=True)

    if args.watch:
        incomplete_stack: set[str] = set()
        failed_stack: set[str] = set()
        empty_stack: set[str] = set()

        def handler(modified_textures: list[Texture]) -> None:
            nonlocal existing_textures, good_writes

            for texture in modified_textures:
                save_path: Path | None = None

                try:
                    save_path = save_texture(
                        texture,
                        output_dir=args.output_dir,
                        args=args,
                    )

                    good_writes += 1
                    empty_stack.discard(texture.uuid)
                    failed_stack.discard(texture.uuid)
                    incomplete_stack.discard(texture.uuid)
                except TextureEmptyError:
                    empty_stack.add(texture.uuid)
                except FileExistsError:
                    existing_textures += 1
                except TextureIncompleteError:
                    incomplete_stack.add(texture.uuid)
                except OSError as e:
                    failed_stack.add(texture.uuid)

                    if args.output_mode == "debug":
                        print(f"error writing {texture.uuid}: {e}")

                if args.output_mode == "progress":
                    printstr = [f"{good_writes} textures extracted"]

                    if len(incomplete_stack):
                        printstr.append(f"{len(incomplete_stack)} incomplete")

                    if len(failed_stack):
                        printstr.append(f"{len(failed_stack)} failed")

                    if existing_textures:
                        printstr.append(f"{existing_textures} existing skipped")

                    if len(empty_stack):
                        printstr.append(f"{len(empty_stack)} empty skipped")

                    print(", ".join(printstr), end="\r", flush=True)

                if args.output_mode in ("files", "debug") and save_path:
                    print(save_path.resolve())

        observer = cache.watch(handler)

        clear_screen()

        if args.output_mode in ("progress", "debug"):
            print(f"watching for changes in {cache.cache_dir.resolve()}")
            print(f"extracting to {args.output_dir.resolve()}")
            print("")
            print("input ctrl+c or ctrl+d to stop")
            print("")

        with interrupthandler() as h:
            try:
                while observer.is_alive() and not h.interrupted:
                    observer.join(1)
            except KeyboardInterrupt:
                pass
            finally:
                observer.stop()
                observer.join()

            end(
                args=args,
                good_writes=good_writes,
                existing_textures=existing_textures,
                incomplete_textures=len(incomplete_stack),
                error_write_textures=len(failed_stack),
                empty_textures=len(empty_stack),
            )

            sys.exit(130)

    else:
        empty_textures = 0
        error_write_textures = 0
        incomplete_textures = 0

        with interrupthandler() as h:
            for texture in tqdm(
                cache,
                total=cache.header.entry_count,
                desc="extracting textures",
                unit="tex",
                delay=1,
                disable=args.output_mode != "progress",
            ):
                if h.interrupted:
                    # break the loop if the user presses ctrl+c
                    break

                try:
                    save_path = save_texture(
                        texture,
                        output_dir=args.output_dir,
                        args=args,
                    )
                    good_writes += 1

                    if args.output_mode in ("files", "debug"):
                        print(save_path.resolve())

                except TextureEmptyError:
                    empty_textures += 1
                except TextureIncompleteError:
                    incomplete_textures += 1
                except FileExistsError:
                    existing_textures += 1
                except OSError:
                    error_write_textures += 1

            end(
                args=args,
                good_writes=good_writes,
                incomplete_textures=incomplete_textures,
                existing_textures=existing_textures,
                error_write_textures=error_write_textures,
                empty_textures=empty_textures,
            )

            if h.interrupted:
                sys.exit(130)

            if args.output_mode == "files" and good_writes == 0:
                print("warning: no textures were written")
                sys.exit(73)



from datetime import datetime
from io import BytesIO
from pathlib import Path
import struct
from uuid import UUID
from typing import Any, Iterator, Self

from .util import format_bytes

HEADER_STRUCT_FORMAT = "fI32sI"
HEADER_BYTE_COUNT = 44

ENTRY_STRUCT_FORMAT = "16BiiI"
ENTRY_BYTE_COUNT = 28

TEXTURE_CACHE_BYTE_COUNT = 600


class Header:
    version: str
    address_size: int
    encoder: str
    entry_count: int

    def __init__(self, version: str, address_size: int, encoder: str, entry_count: int):
        self.version = version
        self.address_size = address_size
        self.encoder = encoder
        self.entry_count = entry_count

    def __repr__(self) -> str:
        return (
            "<Header "
            f'version="{self.version}", '
            f"address_size={self.address_size}, "
            f'encoder="{self.encoder}", '
            f"entry_count={self.entry_count}>"
        )

    def __iter__(self) -> Iterator[tuple[str, Any]]:
        return iter(self.__dict__.items())

    @classmethod
    def from_texture_entries(cls, texture_entries: BytesIO) -> Self:
        texture_entries.seek(0)
        header = texture_entries.read(HEADER_BYTE_COUNT)
        unpack = struct.unpack(HEADER_STRUCT_FORMAT, header)

        return cls(
            version="%0.2f" % unpack[0],
            address_size=unpack[1],
            encoder=unpack[2].decode("utf-8").replace("\x00", ""),
            entry_count=unpack[3],
        )


class Entry:
    uuid: str
    image_size: int
    body_size: int
    time: datetime

    def __init__(self, uuid: str, image_size: int, body_size: int, time: datetime):
        self.uuid = uuid
        self.image_size = image_size
        self.body_size = body_size
        self.time = time

    def __repr__(self) -> str:
        size = format_bytes(self.image_size) if not self.is_empty else "empty"
        return f"<Entry {self.uuid}, {self.time}, {size}>"

    def __eq__(self, value: object) -> bool:
        return (
            isinstance(value, Entry)
            and self.uuid == value.uuid
            and self.time == value.time
            and self.body_size == value.body_size
        )

    @property
    def is_empty(self) -> bool:
        return self.image_size <= 0

    @classmethod
    def from_bytes(cls, b: bytes) -> Self:
        unpack = struct.unpack(ENTRY_STRUCT_FORMAT, b)

        uuid = str(UUID(int=int.from_bytes(unpack[0:16], byteorder="big")))
        rest = unpack[16:]

        return cls(
            uuid=uuid,
            image_size=rest[0],
            body_size=rest[1],
            time=datetime.fromtimestamp(rest[2]),
        )


def decode_texture_entries(texture_entries: BytesIO, entry_count: int) -> list[Entry]:
    texture_entries.seek(HEADER_BYTE_COUNT)
    entries = []

    for _ in range(entry_count):
        entry_bytes = texture_entries.read(ENTRY_BYTE_COUNT)

        if len(entry_bytes) != ENTRY_BYTE_COUNT:
            raise Exception(f"failed to read entry at {texture_entries.tell()}")

        entries.append(Entry.from_bytes(entry_bytes))

    if len(entries) != entry_count:
        raise Exception(
            f"number of read entries {len(entries)} does not match declared count {entry_count}"
        )

    return entries


def read_texture_cache(texture_cache: BytesIO, n: int) -> bytes:
    try:
        texture_cache.seek(TEXTURE_CACHE_BYTE_COUNT * n)
        return texture_cache.read(TEXTURE_CACHE_BYTE_COUNT)
    except OSError:
        raise Exception(
            f"failed to read from texture cache at {TEXTURE_CACHE_BYTE_COUNT * n}"
        )


def texture_location(cache_dir: Path, uuid: str) -> Path:
    subdir = uuid[0]
    texture_file = uuid + ".texture"

    return cache_dir / subdir / texture_file


def read_texture_body(path: Path) -> bytes:
    if not path.is_file():
        raise FileNotFoundError(f"no texture body at {path}")

    with open(path, "rb") as body_file:
        return body_file.read()



from io import BytesIO
from pathlib import Path
from typing import Any, Callable, Iterator, Optional, TypeVar
from watchdog.observers import Observer
from watchdog.observers.api import BaseObserver
from watchdog.events import (
    PatternMatchingEventHandler,
    DirModifiedEvent,
    FileModifiedEvent,
)

from .core import (
    Header,
    Entry,
    read_texture_cache,
    read_texture_body,
    texture_location,
    decode_texture_entries,
)
from .util import format_bytes

from PIL import Image

T = TypeVar("T")


def loads_bytes_io(p: Path) -> BytesIO:
    return BytesIO(p.read_bytes())


class Texture(Entry):
    index: int
    body_path: Path
    loads: Callable[[], bytes]
    """Open texture as a bytes object"""

    def __init__(
        self,
        *,
        index: int,
        entry: Entry,
        body_path: Path,
        loads: Callable[[], bytes],
    ):
        super().__init__(**entry.__dict__)

        self.index = index
        self.body_path = body_path
        self.loads = loads

    def __repr__(self) -> str:
        size = format_bytes(self.image_size) if not self.is_empty else "empty"
        return f"<Texture {self.uuid}, {self.time}, {size}, is_downloaded={self.is_downloaded()}>"

    def is_downloaded(self) -> bool:
        """Check if the texture file is fully downloaded"""
        return self.fs_size() == self.image_size

    def fs_size(self) -> int:
        """Get the size of the texture file on disk"""
        if self.is_empty:
            return 0

        head_size = self.image_size - self.body_size
        body_size = self.body_path.stat().st_size if self.body_path.is_file() else 0

        return head_size + body_size

    def open_image(self) -> Image.Image:
        """Open texture as a pillow image"""

        b = self.loads()
        return Image.open(BytesIO(b), formats=["jpeg2000"])


class TextureCache:
    cache_dir: Path
    texture_entries_file: BytesIO
    texture_cache_file: BytesIO

    header: Header
    entries: list[Entry] = []
    textures: dict[str, Texture] = {}

    def __init__(self, cache_dir: str | Path):
        self.cache_dir = Path(cache_dir)

        if (
            not self.cache_dir.is_dir()
            or not (self.cache_dir / "texture.entries").exists()
            or not (self.cache_dir / "texture.cache").exists()
        ):
            raise FileNotFoundError("path does not contain a proper texture cache")

        self.refresh()

    def __iter__(self) -> Iterator[Texture]:
        return iter(self.textures.values())

    def __reversed__(self) -> Iterator[Texture]:
        return reversed(self.textures.values())

    def __len__(self) -> int:
        return len(self.textures)

    def __repr__(self) -> str:
        total_size = sum(texture.image_size for texture in self)

        return (
            f"<TextureCache {self.cache_dir.resolve()}, "
            f"{self.header.entry_count} entries, "
            f"{format_bytes(total_size)}>"
        )

    def __get_read_bytes(self, i: int, entry: Entry) -> Callable[[], bytes]:
        def read_bytes() -> bytes:
            head = read_texture_cache(self.texture_cache_file, i)

            if entry.image_size <= 601 and entry.body_size == 0:
                # sometimes the file is smaller than 600 bytes, so using the head is
                # sufficient
                return head
            else:
                path = texture_location(self.cache_dir, entry.uuid)
                body = read_texture_body(path)

                return head + body

        return read_bytes

    def refresh(self) -> Iterator[Texture]:
        old_entry_count = self.header.entry_count if hasattr(self, "header") else 0

        self.texture_entries_file = loads_bytes_io(self.cache_dir / "texture.entries")
        self.texture_cache_file = loads_bytes_io(self.cache_dir / "texture.cache")
        self.header = Header.from_texture_entries(self.texture_entries_file)

        self.entries = decode_texture_entries(
            self.texture_entries_file,
            entry_count=self.header.entry_count,
        )

        if self.header.entry_count < old_entry_count:
            # the cache was cleared
            self.textures = {}

        changed_textures: dict[str, Texture] = {}

        for i, entry in enumerate(self.entries):
            if entry != self.get(entry.uuid, None):
                changed_textures[entry.uuid] = Texture(
                    index=i,
                    entry=entry,
                    loads=self.__get_read_bytes(i, entry),
                    body_path=texture_location(self.cache_dir, entry.uuid),
                )

        self.textures |= changed_textures

        return iter(changed_textures.values())

    def watch(self, handler: Callable[[list[Texture]], Any]) -> BaseObserver:
        """Watch the cache directory for changes and call the handler function on updates."""

        def on_modified(event: DirModifiedEvent | FileModifiedEvent) -> None:
            changed_textures = list(self.refresh())

            if changed_textures:
                handler(changed_textures)

        event_handler = PatternMatchingEventHandler(patterns=["texture.entries"])
        setattr(event_handler, "on_modified", on_modified)

        observer = Observer()
        observer.schedule(event_handler, str(self.cache_dir.resolve()))
        observer.start()

        return observer

    def get(self, uuid: str, default: Optional[T] = None) -> Texture | T:
        return self.textures.get(uuid, default)  # type: ignore



from pathlib import Path
import itertools

os_cache_roots = [
    # posix
    Path.home(),
    # windows
    Path(r"%LocalAppData%"),
    # mac
    Path.home() / "Library/Caches",
]

# macos and windows name their viewer directories the same, but posix has different names
viewer_dirs_posix = [".firestorm_x64/cache", ".alchemynext/cache"]
viewer_dirs_macos_win = ["SecondLife", "Firestorm_x64", "AlchemyNext"]
viewer_dirs = viewer_dirs_posix + viewer_dirs_macos_win


def find_texturecache(path: Path) -> Path | None:
    if not path.exists():
        return None

    # if we find a texture.entries file, we're done
    if path.name == "texturecache" and (path / "texture.entries").exists():
        return path

    # otherwise, recurse into the children with some possible names
    for child in path.iterdir():
        if child.is_dir() and child.name == "texturecache":
            return find_texturecache(child)

    return None


def list_texture_caches() -> list[Path]:
    valid_cache_roots = [path for path in os_cache_roots if path.exists()]

    # find all the possible combinations of cache root and viewer dirs
    viewer_caches = [
        Path.joinpath(root, path)
        for root, path in itertools.product(valid_cache_roots, viewer_dirs)
        if Path.joinpath(root, path).exists()
    ]

    # find the texture cache folder
    texturecaches = [find_texturecache(path) for path in viewer_caches]

    # filter out the paths that don't exist
    texturecaches_non_null = [cache for cache in texturecaches if cache is not None]

    if texturecaches_non_null:
        return texturecaches_non_null
    else:
        raise FileNotFoundError("could not find texture cache")



import signal
from contextlib import ContextDecorator
import sys
from types import FrameType
from typing import Any, Callable, Self, TypeAlias

Handlers: TypeAlias = Callable[[int, FrameType | None], Any] | int | None


class interrupthandler(ContextDecorator):
    interrupted: bool = False
    original_sigint_handler: Handlers
    immediate: bool

    def __init__(self, *, immediate: bool = False) -> None:
        self.original_sigint_handler: Handlers = signal.getsignal(signal.SIGINT)
        self.immediate = immediate

    def __enter__(self) -> Self:
        signal.signal(signal.SIGINT, self.handler)

        return self

    def __exit__(self, *exc: Any) -> None:
        signal.signal(signal.SIGINT, self.original_sigint_handler)

        if self.interrupted:
            sys.exit(130)

    def handler(self, signalnum: int, frame: FrameType | None) -> None:
        self.interrupted = True

        if self.immediate:
            self.__exit__()

        print("\ninterrupt signal received, exiting gracefully...")
